local L = AceLibrary("AceLocale-2.2"):new("RaidBuffStatus")
local options = { 
	type='group',
	args = {
		show = {
			type = 'execute',
			name = 'show',
			desc = L["Show the buff report summary window."],
			usage = "show",
			func = function()
				RaidBuffStatus:DoReport()
				RaidBuffStatus:ShowReportFrame()
			end,
		},
		hide = {
			type = 'execute',
			name = 'hide',
			desc = L["Hide the buff report summary window."],
			usage = "hide",
			func = "HideReportFrame",
		},
		toggle = {
			type = 'execute',
			name = 'toggle',
			desc = L["Hide and show the buff report summary window."],
			usage = "toggle",
			func = function()
				if RaidBuffStatus.frame then
					if RaidBuffStatus.frame:IsVisible() then
						RaidBuffStatus:HideReportFrame()
					else
						RaidBuffStatus:DoReport()
						RaidBuffStatus:ShowReportFrame()
					end
				end
			end,
		},
		report = {
			type = 'execute',
			name = 'report',
			desc = L["Report to /raid or /party who is not buffed to the max."],
			usage = "report",
			func = function()
				RaidBuffStatus:DoReport()
				RaidBuffStatus:ReportToChat(true)
			end,
		},
    	},
}


RaidBuffStatus = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceDB-2.0","AceEvent-2.0","AceDebug-2.0")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local BS = AceLibrary("Babble-Spell-2.2")

local addon = RaidBuffStatus

addon:RegisterChatCommand("/rbs",options)
addon:RegisterChatCommand("/raidbuffstatus",options)
addon:RegisterDB("RaidBuffStatusDB")

--addon:SetDebugging(true)

-- The original lists of flasks and elixirs was taken from BigBrother but has changed and increased since then

local flasks = {
	L["Chromatic Resistance"],
	L["Distilled Wisdom"],
	L["Flask of Fortification"],
	L["Flask of Mighty Restoration"],
	L["Flask of Relentless Assault"],
	L["Supreme Power"],
	L["Flask of the Titans"],
	L["Supreme Power of Shattrath"],
	L["Fortification of Shattrath"],
	L["Mighty Restoration of Shattrath"],
	L["Relentless Assault of Shattrath"],
	L["Flask of Pure Death"],
	L["Flask of Blinding Light"],
	L["Chromatic Wonder"],
	L["Unstable Flask of the Bandit"],
	L["Unstable Flask of the Beast"],
	L["Unstable Flask of the Elder"],
	L["Unstable Flask of the Physician"],
	L["Unstable Flask of the Soldier"],
	L["Unstable Flask of the Sorcerer"],
	L["Flask of Pure Death"],
	L["Pure Death of Shattrath"],
	L["Flask of Blinding Light"],
	L["Blinding Light of Shattrath"],
	L["Mr Pinchy's Blessing"],
}

local belixirs = {
	L["Fel Strength Elixir"],
	L["Onslaught Elixir"],
	L["Major Strength"],
	L["Elixir of Major Strength"],
	L["Major Agility"],
	L["Elixir of Major Agility"],
	L["Elixir of Mastery"],
	L["Major Firepower"],
	L["Elixir of Major Firepower"],
	L["Major Shadow Power"],
	L["Elixir of Major Shadow Power"],
	L["Major Frost Power"],
	L["Elixir of Major Frost Power"],
	L["Adept's Elixir"],
	L["Arcane Elixir"],
	L["Greater Arcane Elixir"],
	L["Healing Power"],
	L["Elixir of Healing Power"],
	L["Elixir of the Mongoose"],
	L["Elixir of Greater Firepower"],
}
local gelixirs = {
	L["Major Armor"],
	L["Elixir of Major Defense"],
	L["Greater Armor"],
	L["Elixir of Superior Defense"],
	L["Greater Mana Regeneration"],
	L["Elixir of Major Mageblood"],
	L["Mana Regeneration"],
	L["Mageblood Potion"],
	L["Earthen Elixir"],
	L["Elixir of Draenic Wisdom"],
	L["Elixir of Major Fortitude"],
	L["Greater Intellect"],
	L["Elixir of Empowerment"],
	L["Empowerment"],
	L["Elixir of Ironskin"],
}

local foods = {
	L["Well Fed"],
	L["\"Well Fed\""],
	L["Enlightened"],
	L["Increased Stamina"],
	L["Rumsey Rum Black Label"],
	L["Electrified"],
}

local fortitude = {
	BS["Power Word: Fortitude"],
	BS["Prayer of Fortitude"],
}

local wild = {
	BS["Mark of the Wild"],
	BS["Gift of the Wild"],
}

local intellect = {
	BS["Arcane Intellect"],
	BS["Arcane Brilliance"],
}

local spirit = {
	BS["Divine Spirit"],
	BS["Prayer of Spirit"],
}

local shadow = {
	BS["Shadow Protection"],
	BS["Prayer of Shadow Protection"],
}

local auras = {
	BS["Crusader Aura"],
	BS["Devotion Aura"],
	BS["Retribution Aura"],
	BS["Concentration Aura"],
	BS["Sanctity Aura"],
	BS["Shadow Resistance Aura"],
	BS["Frost Resistance Aura"],
	BS["Fire Resistance Aura"],
}

local aspects = {
	BS["Aspect of the Monkey"],
	BS["Aspect of the Hawk"],
	BS["Aspect of the Beast"],
	BS["Aspect of the Wild"],
	BS["Aspect of the Viper"],
	BS["Aspect of the Cheetah"],
	BS["Aspect of the Pack"],
}

local badaspects = {
	BS["Aspect of the Cheetah"],
	BS["Aspect of the Pack"],
}

local magearmors = {
	BS["Mage Armor"],
	BS["Frost Armor"],
	BS["Ice Armor"],
	BS["Molten Armor"],
}

local classes = {
	"PRIEST",
	"DRUID",
	"PALADIN",
	"ROGUE",
	"MAGE",
	"WARLOCK",
	"SHAMAN",
	"WARRIOR",
	"HUNTER",
}


local raid = {
	classes = {
		PRIEST = {},
		DRUID = {},
		PALADIN = {},
		ROGUE = {},
		MAGE = {},
		WARLOCK = {},
		SHAMAN = {},
		WARRIOR = {},
		HUNTER = {},
	},
	SpiritTalent = {},
	ClassNumbers = {},
	israid = false,
	isparty = false,
	readid = 0,
	size = 1,
}
raid.reset = function()
	raid.classes = {}
	raid.classes.PRIEST = {}
	raid.classes.DRUID = {}
	raid.classes.PALADIN = {}
	raid.classes.ROGUE = {}
	raid.classes.MAGE = {}
	raid.classes.WARLOCK = {}
	raid.classes.SHAMAN = {}
	raid.classes.WARRIOR = {}
	raid.classes.HUNTER = {}
	raid.SpiritTalent = {}
	raid.ClassNumbers = {}
	raid.israid = false
	raid.isparty = false
	raid.readid = 0
	raid.size = 1
	addon:Debug("raid reset")
end


local specicons = {
	UNKNOWN = "Interface\\Icons\\INV_Misc_QuestionMark",
	Hybrid = "Interface\\Icons\\Spell_Nature_ElementalAbsorption",
}

local classicons = {
	PALADIN = "Interface\\Icons\\Ability_ThunderBolt",
	PRIEST = "Interface\\Icons\\INV_Staff_30",
	DRUID = "Interface\\Icons\\Ability_Druid_Maul",
	MAGE = "Interface\\Icons\\INV_Staff_13",
	ROGUE = "Interface\\Icons\\INV_ThrowingKnife_04",
	WARLOCK = "Interface\\Icons\\Spell_Nature_FaerieFire",
	SHAMAN = "Interface\\Icons\\Spell_Nature_BloodLust",
	WARRIOR = "Interface\\Icons\\INV_Sword_27",
	HUNTER = "Interface\\Icons\\INV_Weapon_Bow_07",
	UNKNOWN = "Interface\\Icons\\INV_Misc_QuestionMark",
}


local tfi = {
	namewidth = 120,
	classwidth = 50,
	specwidth = 50,
	specialisationswidth = 120,
	gap = 2,
	edge = 5,
	inset = 3,
	topedge = 45,
	rowheight = 17,
	rowgap = 0,
	maxrows = 40,
	okbuttonheight = 55,
	rowdata = {},
	rowframes = {},
	buttonsize = 15,
	sort = "class",
	sortorder = 1,
}
tfi.namex = tfi.edge
tfi.classx = tfi.namex + tfi.namewidth + tfi.gap
tfi.specx = tfi.classx + tfi.classwidth + tfi.gap
tfi.specialisationsx = tfi.specx + tfi.specwidth + tfi.gap
tfi.framewidth = tfi.specialisationsx + tfi.specialisationswidth + tfi.edge
tfi.rowwidth = tfi.framewidth - tfi.edge - tfi.edge - tfi.inset - tfi.inset


SP = {
	spirittalent = {
		order = 1000,
		icon = BS:GetSpellIcon(BS["Divine Spirit"]),
		tip = L["Can buff Divine Spirit"],
		callalways = true,
		code = function()
			raid.SpiritTalent = {}
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[14] > 0 then
						table.insert(raid.SpiritTalent, name)
						raid.classes.PRIEST[name].talents.specialisations.spirittalent = true
					end
				end
			end
		end,
	},
	improvedspiritone = {
		order = 990,
		icon = BS:GetSpellIcon(BS["Divine Spirit"]),
		tip = L["Can buff improved Divine Spirit level 1"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[15] > 0 then
						raid.classes.PRIEST[name].talents.specialisations.improvedspiritone = true
					end
				end
			end
		end,
	},
	improvedspirittwo = {
		order = 980,
		icon = BS:GetSpellIcon(BS["Divine Spirit"]),
		tip = L["Can buff improved Divine Spirit level 2"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[15] > 1 then
						raid.classes.PRIEST[name].talents.specialisations.improvedspirittwo = true
					end
				end
			end
		end,
	},
	improvedmotwone = {
		order = 970,
		icon = BS:GetSpellIcon(BS["Mark of the Wild"]),
		tip = L["Can buff improved Mark of the Wild level 1"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 0 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwone = true
					end
				end
			end
		end,
	},
	improvedmotwtwo = {
		order = 960,
		icon = BS:GetSpellIcon(BS["Mark of the Wild"]),
		tip = L["Can buff improved Mark of the Wild level 2"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 1 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwtwo = true
					end
				end
			end
		end,
	},
	improvedmotwthree = {
		order = 950,
		icon = BS:GetSpellIcon(BS["Mark of the Wild"]),
		tip = L["Can buff improved Mark of the Wild level 3"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 2 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwthree = true
					end
				end
			end
		end,
	},
	improvedmotwfour = {
		order = 940,
		icon = BS:GetSpellIcon(BS["Mark of the Wild"]),
		tip = L["Can buff improved Mark of the Wild level 4"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 0 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwfour = true
					end
				end
			end
		end,
	},
	improvedmotwfive = {
		order = 930,
		icon = BS:GetSpellIcon(BS["Mark of the Wild"]),
		tip = L["Can buff improved Mark of the Wild level 5"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.DRUID) do
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[1] > 0 then
						raid.classes.DRUID[name].talents.specialisations.improvedmotwfive = true
					end
				end
			end
		end,
	},

	improvedhealthstoneone = {
		order = 920,
		icon = "Interface\\Icons\\INV_Stone_04",
		tip = L["Improved Health Stone level 1"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARLOCK) do
				if raid.classes.WARLOCK[name].talents then
					if raid.classes.WARLOCK[name].talents.tree[2].talent[1] > 0 then
						raid.classes.WARLOCK[name].talents.specialisations.improvedhealthstoneone = true
					end
				end
			end
		end,
	},
	improvedhealthstonetwo = {
		order = 920,
		icon = "Interface\\Icons\\INV_Stone_04",
		tip = L["Improved Health Stone level 2"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.WARLOCK) do
				if raid.classes.WARLOCK[name].talents then
					if raid.classes.WARLOCK[name].talents.tree[2].talent[1] > 1 then
						raid.classes.WARLOCK[name].talents.specialisations.improvedhealthstonetwo = true
					end
				end
			end
		end,
	},
	lightwell = {
		order = 910,
		icon = BS:GetSpellIcon(BS["Lightwell"]),
		tip = L["Can create a Lightwell"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[2].talent[18] > 0 then
						raid.classes.PRIEST[name].talents.specialisations.lightwell = true
					end
				end
			end
		end,
	},
	improvedfortitudeone = {
		order = 900,
		icon = BS:GetSpellIcon(BS["Power Word: Fortitude"]),
		tip = L["Can buff improved Power Word: Fortitude level 1"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[4] > 0 then
						raid.classes.PRIEST[name].talents.specialisations.improvedfortitudeone = true
					end
				end
			end
		end,
	},
	improvedfortitudetwo = {
		order = 890,
		icon = BS:GetSpellIcon(BS["Power Word: Fortitude"]),
		tip = L["Can buff improved Power Word: Fortitude level 2"],
		callalways = false,
		code = function()
			for name,_ in pairs(raid.classes.PRIEST) do
				if raid.classes.PRIEST[name].talents then
					if raid.classes.PRIEST[name].talents.tree[1].talent[4] > 1 then
						raid.classes.PRIEST[name].talents.specialisations.improvedfortitudetwo = true
					end
				end
			end
		end,
	},
}


BF = {
	pvp = {											-- button name
		order = 1000,
		list = "pvplist",								-- list name
		check = "checkpvp",								-- check name
		default = true,									-- default state enabled
		defaultbuff = false,								-- default state report as buff missing
		defaultwarning = true,								-- default state report as warning
		defaultdash = true,								-- default state show on dash
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true,								-- check when unit is not in this zone
		chat = L["<PVP On>: "],								-- chat report
		pre = nil,
		main = function(self, name, class, unit, raid, report)				-- called in main loop
			if UnitIsPVP(unit.unitid) then
				table.insert(report.pvplist, name)
				addon:Debug("is pvp")
			end
		end,
		post = nil,									-- called after main loop
		icon = "Interface\\Icons\\INV_BannerPVP_02",					-- icon
		update = function(self)								-- icon text
			addon:DefaultButtonUpdate(self, report.pvplist, addon.db.profile.checkpvp, true)
		end,
		click = function(self, button, down)						-- button click
			addon:ButtonClick(self, button, down, "pvp")
		end,
		tip = function(self)								-- tool tip
			addon:Tooltip(self, L["PVP is On"], report.pvplist)
		end,
	},
	crusader = {
		order = 990,
		list = "crusaderlist",
		check = "checkcrusader",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Crusader Aura>: "],
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.crusader = true
				if Aura:UnitHasBuff(unit.unitid, BS["Crusader Aura"]) then
					table.insert(report.crusaderlist, name)
					addon:Debug("crusader")
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Crusader Aura"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.crusaderlist, addon.db.profile.checkcrusader, report.checking.crusader or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "crusader")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Loladin has Crusader Aura"], report.crusaderlist)
		end,
	},

	shadows = {
		order = 980,
		list = "shadowslist",
		check = "checkshadows",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Shadow Resistance Aura AND Shadow Protection>: "],
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PRIEST > 0 then
				if class == "PALADIN" then
					report.checking.shadows = true
					if Aura:UnitHasBuff(unit.unitid, BS["Shadow Resistance Aura"]) then
						table.insert(report.shadowslist, name)
						addon:Debug("2 shadow")
					end
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Shadow Resistance Aura"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.shadowslist, addon.db.profile.checkshadows, report.checking.shadows or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "shadows")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Paladin has Shadow Resistance Aura AND Shadow Protection"], report.shadowslist)
		end,
	},

	health = {
		order = 970,
		list = "healthlist",
		check = "checkhealth",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Health less than 80%%>: "],
		main = function(self, name, class, unit, raid, report)
			if not unit.isdead then
				local maxhealth = UnitHealthMax(unit.unitid)
				local health = UnitHealth(unit.unitid)
				local percent = (health/maxhealth)*100
				if percent < 80 then
					table.insert(report.healthlist, name)
					addon:Debug("health")
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_131",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.healthlist, addon.db.profile.checkhealth, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "health")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player has health less than 80%"], report.healthlist)
		end,
	},

	mana = {
		order = 960,
		list = "manalist",
		check = "checkmana",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Mana less than 80%%>: "],
		main = function(self, name, class, unit, raid, report)
			if not unit.isdead then
				if class ~= "WARRIOR" and class ~= "ROGUE" then
					local maxmana = UnitManaMax(unit.unitid)
					local mana = UnitMana(unit.unitid)
					local percent = (mana/maxmana)*100
					if percent < 80 then
						table.insert(report.manalist, name)
						addon:Debug("mana")
					end
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_137",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.manalist, addon.db.profile.checkmana, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "mana")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player has mana less than 80%"], report.manalist)
		end,
	},
	zone = {
		order = 950,
		list = "zonelist",
		check = "checkzone",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true, -- actually has no effect
		chat = L["<Different Zone>: "],
		main = nil, -- done by main code
		post = nil,
		icon = "Interface\\Icons\\INV_Misc_QuestionMark",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.zonelist, addon.db.profile.checkzone, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "zone")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is in a different zone"], report.zonelist)
		end,
	},

	offline = {
		order = 940,
		list = "offlinelist",
		check = "checkoffline",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true, -- actualy has no effect
		chat = L["<Offline>: "],
		main = nil, -- done by main code
		post = nil,
		icon = "Interface\\Icons\\INV_Gizmo_FelStabilizer",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.offlinelist, addon.db.profile.checkoffline, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "offline")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is Offline"], report.offlinelist)
		end,
	},

	afk = {
		order = 930,
		list = "afklist",
		check = "checkoffline",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true,
		chat = L["<AFK>: "],
		main = function(self, name, class, unit, raid, report)
			if UnitIsAFK(unit.unitid) then
				table.insert(report.afklist, name)
				addon:Debug("afk")
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\Trade_Fishing",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.afklist, addon.db.profile.checkafk, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "afk")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is AFK"], report.afklist)
		end,
	},

	dead = {
		order = 920,
		list = "deadlist",
		check = "checkdead",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = true,
		chat = L["<Dead>: "],
		main = function(self, name, class, unit, raid, report)
			if unit.isdead then
				table.insert(report.deadlist, name)
				addon:Debug("dead")
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\Spell_Holy_SenseUndead",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.deadlist, addon.db.profile.checkdead, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "dead")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Player is Dead"], report.deadlist)
		end,
	},


	auras = {
		order = 910,
		list = "auraslist",
		check = "checkauras",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = function(report, raid)
			for _, v in ipairs(report.auraslist) do
				local palalist = {}
				for i, k in pairs(raid.classes.PALADIN) do
					if k.group == v then
						table.insert(palalist, i)
					end
				end
				addon:Say(L["<Paladin Different Aura>: Group "] .. v .. ": " .. table.concat(palalist, ", "))
			end
		end,
		pre = function(self, raid, report)
			report.palapergrouplist = {{},{},{},{},{},{},{},{}}
			report.palaaurapergrouplist = {{},{},{},{},{},{},{},{}}
		end,
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.auras = true
				table.insert(report.palapergrouplist[unit.group], name)
				for _, v in ipairs(auras) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						local foundauraingroup = false
						for _, k in ipairs(report.palaaurapergrouplist[unit.group]) do
							if v == k then
								foundauraingroup = true
								break
							end
						end
						if not foundauraingroup then
							table.insert(report.palaaurapergrouplist[unit.group], v)
						end
					end
				end
			end
		end,
		post = function(self, raid, report)
			for i, v in ipairs(report.palapergrouplist) do
			if # v > 1 then
				if # v > # report.palaaurapergrouplist[i] then
					table.insert(report.auraslist, i)
					addon:Debug("not enough aura:" .. i)
				end
			end
		end
		end,
		icon = BS:GetSpellIcon(BS["Retribution Aura"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.auraslist, addon.db.profile.checkauras, report.checking.auras or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "auras")
		end,
		tip = function(self)
			addon:Tooltip(self, L["There are more Paladins than different Auras in group"], report.auraslist)
		end,
	},

	cheetahpack = {
		order = 900,
		list = "cheetahpacklist",
		check = "checkcheetahpack",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Aspect Cheetah/Pack On>: "],
		main = function(self, name, class, unit, raid, report)
			if class == "HUNTER" then
				report.checking.cheetahpack = true
				local hasbuff = false
				for _, v in ipairs(badaspects) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						hasbuff = true
						break
					end
				end
				if hasbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.cheetahpacklist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.cheetahpacklist, name)
					end
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.cheetahpacklist)
		end,
		icon = BS:GetSpellIcon(BS["Aspect of the Cheetah"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.cheetahpacklist, addon.db.profile.checkcheetahpack, report.checking.cheetahpack or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "cheetahpack")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Aspect of the Cheetah or Pack is on"], report.cheetahpacklist)
		end,
	},

	righteousfury = {
		order = 890,
		list = "righteousfurylist",
		check = "checkrighteousfury",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Righteous Fury>: "],
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.righteousfury = true
				if raid.classes.PALADIN[name].talents then
					if raid.classes.PALADIN[name].talents.spec == "Protection" then
						if not Aura:UnitHasBuff(unit.unitid, BS["Righteous Fury"]) then
							table.insert(report.righteousfurylist, name)
						end
					end
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Righteous Fury"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.righteousfurylist, addon.db.profile.checkrighteousfury, report.checking.righteousfury or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "righteousfury")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Protection paladin with no Righteous Fury"], report.righteousfurylist)
		end,
	},

	soulstone = {
		order = 880,
		list = "nosoulstonelist",
		check = "checksoulstone",
		default = true,
		defaultbuff = false,
		defaultwarning = true,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = true,
		chat = function(report, raid)
			if report.checking.soulstone then
				if # report.soulstonelist < 1 then
					addon:Say(L["<No Soulstone detected>"])
				end
			end
		end,
		pre = function(self, raid, report)
			report.soulstonelist = {}
		end,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.WARLOCK > 0 then
				report.checking.soulstone = true
				if Aura:UnitHasBuff(unit.unitid, BS["Soulstone Resurrection"]) then
					table.insert(report.soulstonelist, name)
				end
			end
		end,
		post = function(self, raid, report)
			if raid.ClassNumbers.WARLOCK > 0 then
				if # report.soulstonelist < 1 then
					table.insert(report.nosoulstonelist, "raid")
				end
			end
		end,
		icon = "Interface\\Icons\\Spell_Shadow_SoulGem",
		update = function(self)
			if addon.db.profile.checksoulstone then
				if report.checking.soulstone then
					if # report.soulstonelist > 0 then
						self.count:SetText("0")
					else
						self.count:SetText("1")
					end
				else
					self.count:SetText("")
				end
			else
				self.count:SetText(".")
			end
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "soulstone")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Someone has a Soulstone"], report.soulstonelist)
		end
	},

	food = {
		order = 500,
		list = "foodlist",
		check = "checkfood",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Well Fed>: "],
		main = function(self, name, class, unit, raid, report)
			local missingbuff = true
			for _, v in ipairs(foods) do
				if Aura:UnitHasBuff(unit.unitid, v) then
					missingbuff = false
					break
				end
			end
			if missingbuff then
				table.insert(report.foodlist, name)
				addon:Debug("food")
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Misc_Food_74",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.foodlist, addon.db.profile.checkfood, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "food")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Not Well Fed"], report.foodlist)
		end
	},
	flask = {
		order = 490,
		list = "flasklist",
		check = "checkflaskir",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		chat = L["<Flask or two Elixirs>: "],
		pre = function(self, raid, report)
			report.belixirlist = {}
			report.gelixirlist = {}
		end,
		main = function(self, name, class, unit, raid, report)
			local missingbuff = true
			for _, v in ipairs(flasks) do
				if Aura:UnitHasBuff(unit.unitid, v) then
					missingbuff = false
					break
				end
			end
			if missingbuff then
				local numbbelixir = 0
				local numbgelixir = 0
				for _, v in ipairs(belixirs) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						numbbelixir = 1
						break
					end
				end
				for _, v in ipairs(gelixirs) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						numbgelixir = 1
						break
					end
				end
				local totalelixir = numbbelixir + numbgelixir
				if totalelixir == 0 then
					table.insert(report.flasklist, name) -- no flask or elixir
					addon:Debug("flask/elix")
				elseif totalelixir == 1 then
					addon:Debug("elix")
					if numbbelixir == 0 then
						table.insert(report.belixirlist, name)
						addon:Debug("belix")
					else
						table.insert(report.gelixirlist, name)
						addon:Debug("gelix")
					end
				end
			end
		end,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_119",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.flasklist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flask")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a Flask or two Elixirs"], report.flasklist)
		end,
	},
	belixir = {
		order = 480,
		list = "belixirlist",
		check = "checkflaskir",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		chat = L["<Battle Elixir>: "],
		pre = nil,
		main = nil,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_111",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.belixirlist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flask")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a Battle Elixir"], report.belixirlist)
		end,
	},
	
	gelixir = {
		order = 470,
		list = "gelixirlist",
		check = "checkflaskir",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = false,
		checkzonedout = false,
		chat = L["<Guardian Elixir>: "],
		pre = nil,
		main = nil,
		post = nil,
		icon = "Interface\\Icons\\INV_Potion_158",
		update = function(self)
			addon:DefaultButtonUpdate(self, report.gelixirlist, addon.db.profile.checkflaskir, true)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "flask")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing a Guardian Elixir"], report.gelixirlist)
		end,
	},

	spirit = {
		order = 460,
		list = "spiritlist",
		check = "checkspirit",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Divine Spirit>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if # raid.SpiritTalent > 0 then
				report.checking.spirit = true
				if class ~= "ROGUE" and class ~= "WARRIOR" then
					local missingbuff = true
					for _, v in ipairs(spirit) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						if addon.db.profile.ReportGroupNumber then
							table.insert(report.spiritlist, name .. "(" .. unit.group .. ")" )
						else
							table.insert(report.spiritlist, name)
						end
						addon:Debug("spirit")
					end
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.spiritlist)
		end,
		icon = BS:GetSpellIcon(BS["Divine Spirit"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.spiritlist, addon.db.profile.checkspirit, report.checking.spirit or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "spirit")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Divine Spirit"], report.spiritlist)
		end,
	},

	intellect = {
		order = 450,
		list = "intellectlist",
		check = "checkintellect",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Arcane Intellect>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.MAGE > 0 then
				report.checking.intellect = true
				if class ~= "ROGUE" and class ~= "WARRIOR" then
					local missingbuff = true
					for _, v in ipairs(intellect) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						if addon.db.profile.ReportGroupNumber then
							table.insert(report.intellectlist, name .. "(" .. unit.group .. ")" )
						else
							table.insert(report.intellectlist, name)
						end
						addon:Debug("int")
					end
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.intellectlist)
		end,
		icon = BS:GetSpellIcon(BS["Arcane Intellect"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.intellectlist, addon.db.profile.checkintellect, report.checking.intellect or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "intellect")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Arcane Intellect"], report.intellectlist)
		end
	},

	wild = {
		order = 440,
		list = "wildlist",
		check = "checkwild",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Mark of the Wild>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.DRUID > 0 then
				report.checking.wild = true
				local missingbuff = true
				for _, v in ipairs(wild) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.wildlist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.wildlist, name)
					end
					addon:Debug("wild")
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.wildlist)
		end,
		icon = BS:GetSpellIcon(BS["Mark of the Wild"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.wildlist, addon.db.profile.checkwild, report.checking.wild or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "wild")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Mark of the Wild"], report.wildlist)
		end,
	},

	fortitude = {
		order = 430,
		list = "fortitudelist",
		check = "checkfortitude",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Power Word: Fortitude>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PRIEST > 0 then
				report.checking.fortitude = true
				local missingbuff = true
				for _, v in ipairs(fortitude) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.fortitudelist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.fortitudelist, name)
					end
					addon:Debug("fortitude")
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.fortitudelist)
		end,
		icon = BS:GetSpellIcon(BS["Power Word: Fortitude"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.fortitudelist, addon.db.profile.checkfortitude, report.checking.fortitude or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "fortitude")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Power Word: Fortitude"], report.fortitudelist)
		end,
	},

	shadow = {
		order = 420,
		list = "shadowlist",
		check = "checkshadow",
		default = false,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Shadow Protection>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if raid.ClassNumbers.PRIEST > 0 then
				report.checking.shadow = true
				local missingbuff = true
				for _, v in ipairs(shadow) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					if addon.db.profile.ReportGroupNumber then
						table.insert(report.shadowlist, name .. "(" .. unit.group .. ")" )
					else
						table.insert(report.shadowlist, name)
					end
					addon:Debug("shadow")
				end
			end
		end,
		post = function(self, raid, report)
			addon:SortNameBySuffix(report.shadowlist)
		end,
		icon = BS:GetSpellIcon(BS["Shadow Protection"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.shadowlist, addon.db.profile.checkshadow, report.checking.shadow or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "shadow")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Missing Shadow Protection"], report.shadowlist)
		end,
	},

	noaura = {
		order = 410,
		list = "noauralist",
		check = "checknoaura",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Paladin Aura>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if class == "PALADIN" then
				report.checking.noaura = true
				local missingbuff = true
				for _, v in ipairs(auras) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.noauralist, name)
					addon:Debug("no aura")
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Devotion Aura"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.noauralist, addon.db.profile.checknoaura, report.checking.noaura or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "noaura")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Paladin has no Aura at all"], report.noauralist)
		end,
	},

	noaspect = {
		order = 400,
		list = "noaspectlist",
		check = "checknoaspect",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Hunter Aspect>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if class == "HUNTER" then
				report.checking.noaspect = true
				local missingbuff = true
				for _, v in ipairs(aspects) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.noaspectlist, name)
					addon:Debug("no aspect")
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Aspect of the Monkey"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.noaspectlist, addon.db.profile.checknoaspect, report.checking.noaspect or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "noaspect")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Hunter has no aspect at all"], report.noaspectlist)
		end,
	},

	innerfire = {
		order = 390,
		list = "innerfirelist",
		check = "checkinnerfire",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Inner Fire>: "],
		main = function(self, name, class, unit, raid, report)
			if class == "PRIEST" then
				report.checking.innerfire = true
				if not Aura:UnitHasBuff(unit.unitid, BS["Inner Fire"]) then
					table.insert(report.innerfirelist, name)
					addon:Debug("innerfire")
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Inner Fire"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.innerfirelist, addon.db.profile.checkinnerfire, report.checking.innerfire or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "innerfire")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Priest is missing Inner Fire"], report.innerfirelist)
		end,
	},

	felarmor = {
		order = 380,
		list = "felarmorlist",
		check = "checkfelarmor",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Fel Armor>: "],
		main = function(self, name, class, unit, raid, report)
			if class == "WARLOCK" then
				report.checking.felarmor = true
				if not Aura:UnitHasBuff(unit.unitid, BS["Fel Armor"]) then
					table.insert(report.felarmorlist, name)
					addon:Debug("felarmor")
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Fel Armor"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.felarmorlist, addon.db.profile.checkfelarmor, report.checking.felarmor or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "felarmor")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Warlock is missing Fel Armor"], report.felarmorlist)
		end,
	},

		magearmor = {
		order = 370,
		list = "magearmorlist",
		check = "checkmagearmor",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Mage Armor>: "],
		pre = nil,
		main = function(self, name, class, unit, raid, report)
			if class == "MAGE" then
				report.checking.magearmor = true
				local missingbuff = true
				for _, v in ipairs(magearmors) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.magearmorlist, name)
					addon:Debug("magearmor")
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Mage Armor"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.magearmorlist, addon.db.profile.checkmagearmor, report.checking.magearmor or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "magearmor")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Mage is missing a Mage Armor"], report.magearmorlist)
		end,
	},

	omenofclarity = {
		order = 360,
		list = "omenofclaritylist",
		check = "checkomenofclarity",
		default = true,
		defaultbuff = true,
		defaultwarning = false,
		defaultdash = true,
		defaultboss = true,
		defaulttrash = true,
		checkzonedout = false,
		chat = L["<Omen of Clarity>: "],
		main = function(self, name, class, unit, raid, report)
			if class == "DRUID" then
				if raid.classes.DRUID[name].talents then
					if raid.classes.DRUID[name].talents.tree[3].talent[8] > 0 then
						report.checking.omenofclarity = true
						if not Aura:UnitHasBuff(unit.unitid, BS["Omen of Clarity"]) then
							table.insert(report.omenofclaritylist, name)
							addon:Debug("omenofclarity")
						end
					end
				end
			end
		end,
		post = nil,
		icon = BS:GetSpellIcon(BS["Omen of Clarity"]),
		update = function(self)
			addon:DefaultButtonUpdate(self, report.omenofclaritylist, addon.db.profile.checkomenofclarity, report.checking.omenofclarity or false)
		end,
		click = function(self, button, down)
			addon:ButtonClick(self, button, down, "omenofclarity")
		end,
		tip = function(self)
			addon:Tooltip(self, L["Druid is missing Omen of Clarity"], report.omenofclaritylist)
		end,
	},
}


report = {
	checking = {},
}
report.reset = function()
	for reportname,_ in pairs(report) do
		if type(report[reportname]) == "number" then
			report[reportname] = 0
		elseif type(report[reportname]) == "table" then
			report[reportname] = {}
		end
	end
	addon:Debug("report reset")
end


local profilelist = {
	ReportSelf = false,
	ReportChat = true,
	ReportGroupNumber = true,
	LockWindow = false,
	IgnoreLastThreeGroups = true,
	x = 0,
	y = 0,
}

for buffcheck, _ in pairs(BF) do
	if BF[buffcheck].list then
		report[BF[buffcheck].list] = {} -- add empty list to report
	end
	if BF[buffcheck].default then  -- if default setting for buff check is enabled
		profilelist[BF[buffcheck].check] = true
	else
		profilelist[BF[buffcheck].check] = false
	end
	for _, defname in ipairs({"buff", "warning", "dash", "boss", "trash"}) do
		if BF[buffcheck]["default" .. defname] then
			profilelist[buffcheck .. defname] = true
		else
			profilelist[buffcheck .. defname] = false
		end
	end
end


local inspectqueuename = ""
local inspectqueueunitid = ""
local inspectqueueclass = ""
local inspectqueuetime = 0
local buttons = {}

addon:RegisterDefaults("profile", profilelist)


-- End of inits


function addon:ShowReportFrame()
	ShowUIPanel(RBSFrame)
end

function addon:HideReportFrame()
	HideUIPanel(RBSFrame)
end

function addon:ToggleTalentsFrame()
	if addon.talentframe:IsVisible() then
		HideUIPanel(TalentsFrame)
	else
		addon:ShowTalentsFrame()
	end
end

function addon:ShowTalentsFrame()
	addon:UpdateTalentsFrame()
	ShowUIPanel(TalentsFrame)
end

function addon:UpdateTalentsFrame()
	local height = tfi.topedge + (raid.size * (tfi.rowheight + tfi.rowgap)) + tfi.okbuttonheight
	addon.talentframe:SetHeight(height)
	for i = 1, tfi.maxrows do
		tfi.rowframes[i].rowframe:Hide()
	end
	for speccheck, _ in pairs(SP) do
		SP[speccheck].code()
	end
	addon:GetTalentRowData()
	addon:SortTalentRowData(tfi.sort, tfi.sortorder)
	addon:CopyTalentRowDataToRowFrames()
	for i = 1, raid.size do
		tfi.rowframes[i].rowframe:Show()
	end
end

function addon:GetTalentRowData()
	tfi.rowdata = {}
	local row = 1
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
			local unit = raid.classes[class][name]
			tfi.rowdata[row] = {}
			tfi.rowdata[row].name = name
			tfi.rowdata[row].class = class
			tfi.rowdata[row].specialisations = {}
			if unit.talents then
				tfi.rowdata[row].spec = unit.talents.spec
				tfi.rowdata[row].specicon = unit.talents.specicon
--				addon:Debug(unit.talents.specicon)
				for speccheck, _ in pairs(SP) do
					if unit.talents.specialisations[speccheck] then
						table.insert(tfi.rowdata[row].specialisations, speccheck)
					end
				end
			else
				tfi.rowdata[row].spec = "UNKNOWN"
				tfi.rowdata[row].specicon = "UNKNOWN"
			end
			
			table.sort(tfi.rowdata[row].specialisations, function (a,b)
				return(SP[a].order > SP[b].order)
			end)
			row = row + 1
		end
	end
end

function addon:SortTalentRowData(sort, sortorder)
	tfi.sort = sort
	tfi.sortorder = sortorder
	if sort == "name" then
		table.sort(tfi.rowdata, function (a,b)
			return (addon:Compare(a.name, b.name, sortorder))
		end)
	elseif sort == "class" then
		table.sort(tfi.rowdata, function (a,b)
			if a.class == b.class then
				return (addon:Compare(a.spec, b.spec, sortorder))
			else
				return (addon:Compare(a.class, b.class, sortorder))
			end
		end)
	elseif sort == "spec" then
		table.sort(tfi.rowdata, function (a,b)
			if a.spec == b.spec then
				return (addon:Compare(a.class, b.class, sortorder))
			else
				return (addon:Compare(a.spec, b.spec, sortorder))
			end
		end)
	elseif sort == "specialisations" then
		table.sort(tfi.rowdata, function (a,b)
			return (addon:Compare(#a.specialisations, #b.specialisations, sortorder))
		end)
	end
end

function addon:Compare(a, b, sortorder)
	if sortorder == 1 then
			return (a < b)
		else
			return (a > b)
	end
end

function addon:CopyTalentRowDataToRowFrames()
	for i, _ in ipairs(tfi.rowdata) do
		local class = tfi.rowdata[i].class
		local name = tfi.rowdata[i].name
		local r = RAID_CLASS_COLORS[class].r
		local g = RAID_CLASS_COLORS[class].g
		local b = RAID_CLASS_COLORS[class].b
		tfi.rowframes[i].name:SetText(name)
		tfi.rowframes[i].name:SetTextColor(r,g,b)
		tfi.rowframes[i].class:SetNormalTexture(classicons[class] or classicons.UNKNOWN)
		tfi.rowframes[i].class:SetScript("OnEnter", function() addon:Tooltip(tfi.rowframes[i].class, addon:TitleCaps(class), nil) end )
		tfi.rowframes[i].class:SetScript("OnLeave", function() GameTooltip:Hide() end)
		if raid.classes[class][name].talents then
			local spec = raid.classes[class][name].talents.spec
			local specicon = raid.classes[class][name].talents.specicon
			tfi.rowframes[i].spec:SetScript("OnEnter", function() addon:Tooltip(tfi.rowframes[i].spec, spec) end )
			tfi.rowframes[i].spec:SetScript("OnLeave", function() GameTooltip:Hide() end)
			if spec == "Hybrid" then
				tfi.rowframes[i].spec:SetNormalTexture(specicons.Hybrid)
			else
				tfi.rowframes[i].spec:SetNormalTexture(specicon or specicons.UNKNOWN)
			end
		else
			tfi.rowframes[i].spec:SetNormalTexture(specicons.UNKNOWN)
			tfi.rowframes[i].spec:SetScript("OnEnter", function() addon:Tooltip(tfi.rowframes[i].spec, "Unknown") end )
			tfi.rowframes[i].spec:SetScript("OnLeave", function() GameTooltip:Hide() end)
		end
		for j, v in ipairs (tfi.rowframes[i].specialisations) do
			v:Hide()
			local speccheck = tfi.rowdata[i].specialisations[j]
			if speccheck then
				v:SetNormalTexture(SP[speccheck].icon)
				v:SetScript("OnEnter", function() addon:Tooltip(v, SP[speccheck].tip) end )
				v:SetScript("OnLeave", function() GameTooltip:Hide() end)
				v:Show()
			end
		end
	end
end

function addon:DoReport()
	doingreport = true
	report:reset()
	addon:ReadRaid()
	if (not raid.israid) and (not raid.isparty) then
		addon:Debug("Report aborted as not in raid or party.")
		return
	end
	addon:CalculateReport()
	addon:UpdateButtons()
end

function addon:CalculateReport()
	-- PRE HERE
	for buffcheck, _ in pairs(BF) do
		if addon.db.profile[BF[buffcheck].check] then
			if BF[buffcheck].pre then
--				addon:Debug("pre:" .. buffcheck)
				BF[buffcheck].pre(self, raid, report)
			end
		end
	end

	-- MAIN HERE
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
--			addon:Debug("Name:" .. name .. " " .. class)
			local unit = raid.classes[class][name]

			if not unit.online then
				if addon.db.profile.checkoffline then
					table.insert(report.offlinelist, name)  -- used by offline warning check
				end
				break
			end

			local zonedin = true
			if raid.israid then
				if GetRealZoneText() ~= unit.zone then
					zonedin = false
					if addon.db.profile.checkzone then
						table.insert(report.zonelist, name)
						addon:Debug("zone")
					end
				end
			end

			for buffcheck, _ in pairs(BF) do
				if addon.db.profile[BF[buffcheck].check] then
					if zonedin or BF[buffcheck].checkzonedout then
						if BF[buffcheck].main then
							BF[buffcheck].main(self, name, class, unit, raid, report)
						end
					end
				end
			end
			if unit.talents == nil then
				addon:RequestInspect(name, class, unit.unitid) -- find out all talent info eventually
			end

		end

	end

	-- POST HERE
	for buffcheck, _ in pairs(BF) do
		if addon.db.profile[BF[buffcheck].check] then
			if BF[buffcheck].post then
--				addon:Debug("post:" .. buffcheck)
				BF[buffcheck].post(self, raid, report)
			end
		end
	end
end

function addon:ReportToChat(boss, player)
	local warnings = 0
	local buffs = 0
	for buffcheck, _ in pairs(BF) do
		if # report[BF[buffcheck].list] > 0 then
			if (boss and addon.db.profile[buffcheck .. "boss"]) or ((not boss) and addon.db.profile[buffcheck .. "trash"]) then
				if addon.db.profile[buffcheck .. "warning"] then
					warnings = warnings + # report[BF[buffcheck].list]
				end
				if addon.db.profile[buffcheck .. "buff"] then
					buffs = buffs + # report[BF[buffcheck].list]
				end
			end
		end
	end
	if warnings > 0 then
		addon:Say(L["Warnings: "] .. warnings)
		for buffcheck, _ in pairs(BF) do
			if # report[BF[buffcheck].list] > 0 then
				if (boss and addon.db.profile[buffcheck .. "boss"] ) or ((not boss) and addon.db.profile[buffcheck .. "trash"]) then
					if addon.db.profile[buffcheck .. "warning"] then
						if type(BF[buffcheck].chat) == "string" then
							addon:Say(BF[buffcheck].chat .. table.concat(report[BF[buffcheck].list], ", "))
						elseif type(BF[buffcheck].chat) == "function" then
							BF[buffcheck].chat(report, raid)
						end
					end
				end
			end
		end
	end
	if buffs > 0 then
		if boss then
			addon:Say(L["Missing buffs (Boss): "] .. buffs)
		else
			addon:Say(L["Missing buffs (Trash): "] .. buffs)
		end
		for buffcheck, _ in pairs(BF) do
			if # report[BF[buffcheck].list] > 0 then
				if (boss and addon.db.profile[buffcheck .. "boss"] ) or ((not boss) and addon.db.profile[buffcheck .. "trash"]) then
					if addon.db.profile[buffcheck .. "buff"] then
						if type(BF[buffcheck].chat) == "string" then
							addon:Say(BF[buffcheck].chat .. table.concat(report[BF[buffcheck].list], ", "))
						elseif type(BF[buffcheck].chat) == "function" then
							BF[buffcheck].chat(report, raid)
						end
					end
				end
			end
		end
	else
		addon:Say(L["No buffs needed!"])
	end

end




function addon:ReadRaid()
	raid.readid = raid.readid + 1
	local raidnum = GetNumRaidMembers()
	local partynum = GetNumPartyMembers()
	if raidnum < 2 then
		if partynum < 1 then
			raid.reset()
			return
		else
			addon:Debug("Party detected.")
			raid.isparty = true
			raid.israid = false
			raid.size = partynum + 1
			for i = 1, partynum do
				addon:ReadUnit("party" .. i)
			end
			addon:ReadUnit("player")
		end
	else
		if raid.isparty then
			addon:Debug("Party has converted to Raid!")
			raid.reset()
		else
			addon:Debug("Raid detected.")
		end
		raid.isparty = false
		raid.israid = true
		raid.size = raidnum
		for i = 1, raidnum do
			addon:ReadUnit("raid" .. i)
		end
	end
	addon:DeleteOldUnits()
	addon:CalculateClassNumbers()
	for speccheck, _ in pairs(SP) do
		if SP[speccheck].callalways then
			SP[speccheck].code()
		end
	end
end


-- raid = { classes = { CLASS = { NAME = { readid, unitid, group, zone, online, isdead, talents = {spec, tree = { talent = {} 
function addon:ReadUnit(unitid)
	if not UnitExists(unitid) then
		addon:Debug("Unit does not exist:" .. unitid)
		return
	end
	local unitindex = select(3, unitid:find("(%d+)"))
	local name = UnitName(unitid)
	if name and name ~= UNKNOWNOBJECT and name ~= UKNOWNBEING then

		local class = select(2, UnitClass(unitid))
		local isDead = UnitIsDeadOrGhost(unitid) or false
		local rank = 0
		local subgroup = 1
		local online =  UnitIsConnected(unitid)
		local role = ""
		local zone = "UNKNOWN"
		local nametwo = name
		local isML = false
		local level = UnitLevel(unitid)
		
		if raid.israid then
			nametwo, rank, subgroup, _, _, _, zone, _, _, role, isML = GetRaidRosterInfo(unitindex)
			if name ~= nametwo then
				addon:Debug("NAMES DIFFER!" .. name .. " " .. nametwo)
			end
		end
		if addon.db.profile.IgnoreLastThreeGroups then
			if subgroup > 5 then
				raid.size = raid.size - 1
				return -- quick hack for now to ignore ppl in grps 6-8
			end
		end
		if raid.classes[class][name] == nil then
			raid.classes[class][name] = {}
		end
		local rcn = raid.classes[class][name]
		rcn.readid = raid.readid
		rcn.unitid = unitid
		rcn.group = subgroup
		rcn.zone = zone
		rcn.online = online
		rcn.isdead = isDead
		rcn.role = role
		rcn.rank = rank
		rcn.level = level
	end
end

function addon:DeleteOldUnits()
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
			if raid.classes[class][name].readid < raid.readid then
				addon:Debug("Left raid:" .. name)
				raid.classes[class][name] = nil
			end
		end
	end
end

function addon:RequestInspect(name, class, unitid)
	if raid.classes[class][name].level < 10 then
--		addon:Debug("Inspect skipping: " .. name .. " " .. raid.classes[class][name].level);
		return -- no talents below lvl 10
	end
	if _G.InspectFrame and _G.InspectFrame:IsShown() then
		return -- can't inspect at same time as UI
	end
	if CanInspect(unitid) and inspectqueuetime < GetTime() then
		if inspectqueuename then
			if inspectqueuename == name then
				inspectqueuename = nil
				inspectqueueclass = nil
				inspectqueueunitid = nil
				return -- must be a timed out retry so skip and allow next request to get in the queue
			end
		end
	 	addon:Debug("Inspect request going to be sent for: " .. name .. " " .. class .. " " .. unitid);
		inspectqueuetime = GetTime() + 15
		inspectqueuename = name
		inspectqueueunitid = unitid
		inspectqueueclass = class
		if UnitIsUnit(inspectqueueunitid, "player") then
			addon:INSPECT_TALENT_READY()
		else
			NotifyInspect(unitid)
			addon:RegisterEvent("INSPECT_TALENT_READY")
		end
	end
end


function addon:INSPECT_TALENT_READY()
	if addon:IsEventRegistered("INSPECT_TALENT_READY") then
		self:UnregisterEvent("INSPECT_TALENT_READY")
	end
	if not inspectqueuename then
		addon:Debug("INSPECT_TALENT_READY: nothing queued");
		inspectqueuetime = 0
		return
	end
	local isnotme = true
	if UnitIsUnit(inspectqueueunitid, "player") then
		addon:Debug("INSPECT_TALENT_READY: player is me");
		isnotme = false
	end
	if GetNumTalentTabs(isnotme) == 0 then -- sometimes the inspect fails
		addon:Debug("Inspect failed: " .. inspectqueuename);
		ClearInspectPlayer()
		inspectqueuetime = 0
		return
	end
	local name = inspectqueuename
	local class = inspectqueueclass
	inspectqueuename = nil
	inspectqueueclass = nil
	raid.classes[class][name].talents = { tree = {}, spec = "UNKNOWN", specicon = "UNKNOWN", specialisations = {} }
	local tree = raid.classes[class][name].talents.tree
	local tabtotal = {0, 0, 0}
	for tab = 1, GetNumTalentTabs(isnotme) do
		local name, icon, spent = GetTalentTabInfo(tab, isnotme);
		tree[tab] = {name=name, icon=icon, spent=spent, talent={ }};
		for talent = 1, GetNumTalents(tab, isnotme) do
			tree[tab].talent[talent] = select(5, GetTalentInfo(tab, talent, isnotme));
			tabtotal[tab] = tabtotal[tab] + tree[tab].talent[talent]
			addon:Debug(tab .. " " .. talent .. " " .. tree[tab].talent[talent])
		end
	end
	if tabtotal[1] > 31 then
		raid.classes[class][name].talents.spec = tree[1].name
		raid.classes[class][name].talents.specicon = tree[1].icon
	elseif tabtotal[2] > 31 then
		raid.classes[class][name].talents.spec = tree[2].name
		raid.classes[class][name].talents.specicon = tree[2].icon
	elseif tabtotal[3] > 31 then
		raid.classes[class][name].talents.spec = tree[3].name
		raid.classes[class][name].talents.specicon = tree[3].icon
	else
		raid.classes[class][name].talents.spec = "Hybrid"
	end
	addon:Debug(class .. " " .. name .. " " .. raid.classes[class][name].talents.spec .. " " .. raid.classes[class][name].talents.specicon )
	addon:Debug("Inspect request finished: " .. name);
	if addon.talentframe:IsVisible() then
		addon:UpdateTalentsFrame()
	end
	inspectqueuetime = 0
end


function addon:CalculateClassNumbers()
	for _,class in ipairs(classes) do
		raid.ClassNumbers[class] = addon:DicSize(raid.classes[class])
	end
end


function addon:Say(fmt, text, player)
	local msg = string.format(fmt, text)
	local str = "RBS::"
	local canspeak = IsRaidLeader() or IsRaidOfficer()
	for _,s in pairs({strsplit(" ", msg)}) do
		if #str + #s >= 250 then
			if player then SendChatMessage(str, player) end
			if addon.db.profile.ReportChat and raid.isparty then SendChatMessage(str, "party") end
			if addon.db.profile.ReportChat and raid.israid and canspeak then SendChatMessage(str, "raid") end
			if addon.db.profile.ReportSelf then addon:Print(str) end
			str = "RBS::"
		end
		str = str .. " " .. s
	end
	if player then SendChatMessage(str, player) end
	if addon.db.profile.ReportChat and raid.isparty then SendChatMessage(str, "party") end
	if addon.db.profile.ReportChat and raid.israid and canspeak then SendChatMessage(str, "raid") end
	if addon.db.profile.ReportSelf then addon:Print(str) end
end


function addon:DicSize(dic) -- is there really no built-in function to do this??
	local i = 0
	for _,_ in pairs(dic) do
		i = i + 1
	end
	return i
end


function addon:SetupFrames()
	-- main frame
	addon.frame = CreateFrame("Frame", "RBSFrame", UIParent)
	addon.frame:Hide()
	addon.frame:EnableMouse(true)
	addon.frame:SetFrameStrata("MEDIUM")
	addon.frame:SetMovable(true)
	addon.frame:SetToplevel(true)
	addon.frame:SetWidth(128)
	addon.frame:SetHeight(190)
	addon.rbsfs = addon.frame:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText("RBS " .. GetAddOnMetadata("RaidBuffStatus", "Version"))
	addon.rbsfs:SetPoint("TOP",0,-5)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.frame:SetBackdrop( { 
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	addon.frame:SetBackdropBorderColor(0,0,0,200)
	addon.frame:SetBackdropColor(0,0,0,100)
	addon.frame:ClearAllPoints()
	addon.frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	addon.frame:SetScript("OnMouseDown",function()
		if ( arg1 == "LeftButton" ) then
			if not addon.db.profile.LockWindow then
				this:StartMoving()
			end
		end
	end)
	addon.frame:SetScript("OnMouseUp",function()
		if ( arg1 == "LeftButton" ) then
			this:StopMovingOrSizing()
			addon:SaveFramePosition()
		end
	end)
	addon.frame:SetScript("OnHide",function() this:StopMovingOrSizing() end)
	addon.frame:SetClampedToScreen(true)
	addon:LoadFramePosition()
	addon:AddBuffButtons()
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["Boss"])
	addon.button:SetWidth(54)
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", 10, 5)
	addon.button:SetScript("OnClick", function() addon:DoReport() addon:ReportToChat(true) end)
	addon.button:Show()
	
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["Trash"])
	addon.button:SetWidth(54)
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", 64, 5)
	addon.button:SetScript("OnClick", function() addon:DoReport() addon:ReportToChat(false) end)
	addon.button:Show()

	addon.talentsbutton = CreateFrame("Button", "talentsbutton", addon.frame, "SecureActionButtonTemplate")
	addon.talentsbutton:SetWidth(20)
	addon.talentsbutton:SetHeight(20)
	addon.talentsbutton:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	addon.talentsbutton:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down") 
	addon.talentsbutton:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	addon.talentsbutton:ClearAllPoints()
	addon.talentsbutton:SetPoint("TOPLEFT", addon.frame, "TOPLEFT", 5, -5)
	addon.talentsbutton:SetScript("OnClick", function()
		addon:ToggleTalentsFrame()
	end
	)
	addon.talentsbutton:Show()
	addon.optionsbutton = CreateFrame("Button", "optionsbutton", addon.frame, "SecureActionButtonTemplate")
	addon.optionsbutton:SetWidth(20)
	addon.optionsbutton:SetHeight(20)
	addon.optionsbutton:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	addon.optionsbutton:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down") 
	addon.optionsbutton:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	addon.optionsbutton:ClearAllPoints()
	addon.optionsbutton:SetPoint("TOPRIGHT", addon.frame, "TOPRIGHT", -5, -5)
	addon.optionsbutton:SetScript("OnClick", function()
		addon:Say("<Options Window>: RBS")
	end
	)
	addon.optionsbutton:Show()

	-- talents window frame

	addon.talentframe = CreateFrame("Frame", "TalentsFrame", UIParent, "DialogBoxFrame")
	addon.talentframe:Hide()
	addon.talentframe:EnableMouse(true)
	addon.talentframe:SetFrameStrata("MEDIUM")
	addon.talentframe:SetMovable(true)
	addon.talentframe:SetToplevel(true)
	addon.talentframe:SetWidth(tfi.framewidth)
	addon.talentframe:SetHeight(190)
	addon.rbsfs = addon.talentframe:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText("RBS " .. GetAddOnMetadata("RaidBuffStatus", "Version") .. " - " .. L["Talent Specialisations"])
	addon.rbsfs:SetPoint("TOP",0,-5)
	addon.rbsfs:SetTextColor(.9,0,0)
	addon.rbsfs:Show()
	addon.talentframe:SetBackdrop( { 
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	addon.talentframe:SetBackdropBorderColor(0,0,0,200)
	addon.talentframe:SetBackdropColor(0,0,0,100)
	addon.talentframe:ClearAllPoints()
	addon.talentframe:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	addon.talentframe:SetScript("OnMouseDown",function()
		if ( arg1 == "LeftButton" ) then
			if not addon.db.profile.LockWindow then
				this:StartMoving()
			end
		end
	end)
	addon.talentframe:SetScript("OnMouseUp",function()
		if ( arg1 == "LeftButton" ) then
			this:StopMovingOrSizing()
			addon:SaveFramePosition()
		end
	end)
	addon.talentframe:SetScript("OnHide",function() this:StopMovingOrSizing() end)
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Name"])
	addon.button:SetWidth(tfi.namewidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.namex, -20)
	addon.button:SetScript("OnClick", function()
		tfi.sort = "name"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Class"])
	addon.button:SetWidth(tfi.classwidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.classx, -20)
		addon.button:SetScript("OnClick", function()
		tfi.sort = "class"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Spec"])
	addon.button:SetWidth(tfi.specwidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.specx, -20)
		addon.button:SetScript("OnClick", function()
		tfi.sort = "spec"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()
	addon.button = CreateFrame("Button", "", addon.talentframe, "OptionsButtonTemplate")
	addon.button:SetText(L["Specialisations"])
	addon.button:SetWidth(tfi.specialisationswidth)
	addon.button:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.specialisationsx, -20)
		addon.button:SetScript("OnClick", function()
		tfi.sort = "specialisations"
		tfi.sortorder = 0 - tfi.sortorder
		addon:ShowTalentsFrame()
	end)
	addon.button:Show()

	rowy = 0 - tfi.topedge
	for i = 1, tfi.maxrows do
		tfi.rowframes[i] = {}
		addon.rowframe = CreateFrame("Frame", "", addon.talentframe)
		tfi.rowframes[i].rowframe = addon.rowframe
		addon.rowframe:SetWidth(tfi.rowwidth)
		addon.rowframe:SetHeight(tfi.rowheight)
		addon.rowframe:ClearAllPoints()
		addon.rowframe:SetPoint("TOPLEFT", addon.talentframe, "TOPLEFT", tfi.edge + tfi.inset, rowy)
		addon.rbsfs = addon.rowframe:CreateFontString(nil,"ARTWORK","GameFontNormal")
		tfi.rowframes[i].name = addon.rbsfs
		addon.rbsfs:SetText("Whoami")
		addon.rbsfs:SetPoint("TOPLEFT", addon.rowframe, "TOPLEFT", 0, -2)
		addon.rbsfs:SetTextColor(.9,0,0)
		addon.rbsfs:Show()

		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].class = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\INV_ValentinesCandy")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPLEFT", addon.rowframe, "TOPLEFT", tfi.classx + ((tfi.classwidth - 30) / 2), 0)
		addon.button:Show()

		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].spec = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPLEFT", addon.rowframe, "TOPLEFT", tfi.specx + ((tfi.specwidth - 30) / 2), 0)
		addon.button:Show()
		
		tfi.rowframes[i].specialisations = {}
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[1] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", addon.rowframe, "TOPRIGHT", 0 - tfi.inset, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[2] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[1], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[3] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[2], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[4] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[3], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[5] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[4], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		addon.button = CreateFrame("Button", "", addon.rowframe)
		tfi.rowframes[i].specialisations[6] = addon.button
		addon.button:SetWidth(tfi.buttonsize)
		addon.button:SetHeight(tfi.buttonsize)
		addon.button:SetNormalTexture("Interface\\Icons\\Ability_ThunderBolt")
		addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		addon.button:SetPoint("TOPRIGHT", tfi.rowframes[i].specialisations[5], "TOPLEFT", 0 - tfi.gap, 0)
		addon.button:Show()
		rowy = rowy - tfi.rowheight - tfi.rowgap
	end
end


function addon:SaveFramePosition()
	addon.db.profile.x = addon.frame:GetLeft()
	addon.db.profile.y = addon.frame:GetTop() - UIParent:GetTop()
end

function addon:LoadFramePosition()
	addon.frame:ClearAllPoints()
	if (addon.db.profile.x ~= 0) or (addon.db.profile.y ~= 0) then
		addon.frame:SetPoint("TOPLEFT", UIParent,"TOPLEFT", addon.db.profile.x, addon.db.profile.y)
	else
		addon.frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	end
end


function addon:AddBuffButtons()
	local buffs = {}
	local warnings = {}
	local bosses = {}
	for buffcheck, _ in pairs(BF) do
		if addon.db.profile[buffcheck .. "dash"] then
			if addon.db.profile[buffcheck .. "boss"] and (not addon.db.profile[buffcheck .. "trash"]) then
				table.insert(bosses, buffcheck)
			else
				if addon.db.profile[buffcheck .. "warning"] then
					table.insert(warnings, buffcheck)
				end
				if addon.db.profile[buffcheck .. "buff"] then
					table.insert(buffs, buffcheck)
				end
			end
		end
	end
	addon:SortButtons(bosses)
	addon:SortButtons(buffs)
	addon:SortButtons(warnings)
	local currenty = 8
	local maxcols = 5
	local cols = { 10, 32, 54, 76, 98, 120, 142, 164, 186 }
	if # bosses > 0 then
		currenty = addon:AddButtonType(bosses, maxcols, cols, currenty)
	end
	if # buffs > 0 then
		currenty = addon:AddButtonType(buffs, maxcols, cols, currenty)
	end
	if # warnings > 0 then
		currenty = addon:AddButtonType(warnings, maxcols, cols, currenty)
	end

	addon.frame:SetHeight(currenty + 42)
end

function addon:AddButtonType(buttonlist, maxcols, cols, currenty)
	for i, v in ipairs(buttonlist) do
		local x = cols[((i - 1) % maxcols) + 1]
		local y = currenty + (22 * (math.ceil((# buttonlist)/maxcols) - math.floor((i - 1) / maxcols)))
		addon:AddBuffButton(v, x, y, BF[v].icon, BF[v].update, BF[v].click, BF[v].tip)
	end
	return (currenty + 6 + 22 * math.ceil((# buttonlist)/maxcols))
end

function addon:SortButtons(buttonlist)
	table.sort(buttonlist, function (a,b)
		return (BF[a].order > BF[b].order)
	end)
end


function addon:AddBuffButton(name, x, y, icon, update, click, tooltip)
	addon.button = CreateFrame("Button", name, addon.frame, "SecureActionButtonTemplate")
	table.insert(buttons, addon.button)
	addon.button:Hide()
	addon.button:SetWidth(20)
	addon.button:SetHeight(20)
	addon.button:SetNormalTexture(icon or BS:GetSpellIcon(spell) or "Interface\\Icons\\INV_ValentinesCandy")
	addon.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	addon.count = addon.button:CreateFontString(nil, "ARTWORK","GameFontNormalSmall")
	addon.button.count = addon.count
	addon.count:SetWidth(20)
	addon.count:SetHeight(20)
	addon.count:SetFont(addon.count:GetFont(),12,"OUTLINE")
	addon.count:SetPoint("CENTER", self.button, "CENTER", 0, 0)
	addon.count:SetText("X")
	addon.count:Show()
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", x, y)
	if click then
		addon.button:SetScript("OnClick", click)
	end
	if update then
		addon.button.update = update
	end
	if tooltip then
		addon.button:SetScript("OnEnter", tooltip)
		addon.button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	end
	addon.button:Show()
end

function addon:ToggleCheck(...)
	if addon.db.profile[...] then
		addon.db.profile[...] = false
	else
		addon.db.profile[...] = true
		addon:DoReport()
	end
	addon:UpdateButtons()
end


function addon:UpdateButtons()
	for _,v in ipairs(buttons) do
		v:update(v)
	end
end

function addon:OnProfileEnable()
	addon:Debug("Profile changed")
	addon:LoadFramePosition()
	addon:DoReport()
end

function addon:JoinedPartyRaidChanged()
	addon:Debug("JoinedPartyRaidChanged");
	local oldstatus = raid.isparty or raid.israid
	addon:DoReport()
	if oldstatus then  -- was a raid or party last check
		if raid.isparty or raid.israid then -- still is a raid or party
			addon:Debug("Still in party or raid");
		else	-- no longer in raid or party
			addon:HideReportFrame()
			if addon:IsEventRegistered("PLAYER_AURAS_CHANGED") then
				addon:UnregisterEvent("PLAYER_AURAS_CHANGED")
			end
			if addon:IsEventScheduled("DoReport") then
				addon:CancelScheduledEvent("DoReport")
			end
		end
	else
		if raid.isparty or raid.israid then -- newly entered raid or party
			addon:ShowReportFrame()
			addon:RegisterEvent("PLAYER_AURAS_CHANGED", "DoReport")
			addon:ScheduleRepeatingEvent("DoReport", addon.DoReport, 10, addon)
		end
	end
end

function addon:NotifyInspect(unitid)
	if unitid ~= inspectqueueunitid then
		addon:Debug("NotifyInspect mismatch:" .. unitid .. " " .. inspectqueueunitid)
		if addon:IsEventRegistered("INSPECT_TALENT_READY") then
			self:UnregisterEvent("INSPECT_TALENT_READY")
		end
		inspectqueuename = nil
		inspectqueueunitid = nil
		inspectqueuetime = 0
		return
	end
end

function addon:OnEnable()
--	addon:Print("RBS: Activated");
	addon:SetupFrames()
	hooksecurefunc("NotifyInspect", function(...) return addon:NotifyInspect(...) end)
	addon:JoinedPartyRaidChanged()
	addon:RegisterEvent("RAID_ROSTER_UPDATE", "JoinedPartyRaidChanged")
	addon:RegisterEvent("PARTY_MEMBERS_CHANGED", "JoinedPartyRaidChanged")
	addon:RegisterEvent("PLAYER_ENTERING_WORLD", "JoinedPartyRaidChanged")
end


function addon:OnDisable()
	if addon:IsEventRegistered("RAID_ROSTER_UPDATE") then
		addon:UnregisterEvent("RAID_ROSTER_UPDATE")
	end
	if addon:IsEventRegistered("PARTY_ROSTER_UPDATE") then
		addon:UnregisterEvent("PARTY_ROSTER_UPDATE")
	end
	if addon:IsEventRegistered("PLAYER_ENTERING_WORLD") then
		addon:UnregisterEvent("PLAYER_ENTERING_WORLD")
	end
	if addon:IsEventScheduled("DoReport") then
		addon:CancelScheduledEvent("DoReport")
	end
	if addon:IsEventRegistered("INSPECT_TALENT_READY") then
		self:UnregisterEvent("INSPECT_TALENT_READY")
	end
end

function addon:Tooltip(self, title, list)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
	GameTooltip:SetText(title,1,1,1,1,1)
	if list then
		if #list > 0 then
			local str = ""
			for _,s in pairs({strsplit(" ", table.concat(list, ", "))}) do
				if #str + #s >= 50 then
					GameTooltip:AddLine(str,nil,nil,nil,nil,1)
					str = ""
				end
				str = str .. " " .. s
			end
			GameTooltip:AddLine(str,nil,nil,nil,nil,1)
		end
	end
	GameTooltip:Show()
end

function addon:DefaultButtonUpdate(self, table, profile, checking)
	if profile == false then
		self.count:SetText(".")
	else
		if checking then
			self.count:SetText(# table)
		else
			self.count:SetText("")
		end
	end
end


function addon:ButtonClick(self, button, down, buffcheck)
	local check = BF[buffcheck].check
	if addon.db.profile[check] then
		if IsShiftKeyDown() then
			if type(BF[buffcheck].chat) == "string" then
				if # report[BF[buffcheck].list] > 0 then
					addon:Say(BF[buffcheck].chat .. table.concat(report[BF[buffcheck].list], ", "))
				end
			elseif type(BF[buffcheck].chat) == "function" then
				BF[buffcheck].chat(report, raid)
			end
		else
			addon:ToggleCheck(check)
		end
	else
		addon:ToggleCheck(check)
	end
end

function addon:SortNameBySuffix(thetable)
	table.sort(thetable, function (a,b)
		local grpa = select(3, a:find"(%d+)")
		local grpb = select(3, b:find"(%d+)")
		if grpa and grpb then
			return grpa < grpb
		else
			return a < b
		end
	end)
end

function addon:TitleCaps(str)
	str = string.lower(str)
	return str:gsub("%a", string.upper, 1)
end